/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
import * as pathlib from 'path';
import * as fs from 'fs/promises';
/**
 * Reads package.json files and caches them.
 */
export class CachingPackageJsonReader {
    #cache = new Map();
    async read(packageDir) {
        let packageJson = this.#cache.get(packageDir);
        if (packageJson === undefined) {
            const packageJsonPath = pathlib.resolve(packageDir, 'package.json');
            let packageJsonStr;
            try {
                packageJsonStr = await fs.readFile(packageJsonPath, 'utf8');
            }
            catch (error) {
                if (error.code === 'ENOENT') {
                    throw new CachingPackageJsonReaderError('missing-package-json');
                }
                throw error;
            }
            try {
                packageJson = JSON.parse(packageJsonStr);
            }
            catch (error) {
                throw new CachingPackageJsonReaderError('invalid-package-json');
            }
            this.#cache.set(packageDir, packageJson);
        }
        return packageJson;
    }
}
/**
 * An exception thrown by {@link CachingPackageJsonReader}.
 *
 * Note we don't use {@link WireitError} here because we don't have the full
 * context of the script we're trying to evaluate.
 */
class CachingPackageJsonReaderError extends Error {
    reason;
    constructor(reason) {
        super(reason);
        this.reason = reason;
    }
}
//# sourceMappingURL=package-json-reader.js.map